
#if HAVE_CONFIG_H
    #include <config.h>
#endif

#include <iostream>
#include <fstream>
#include <string>
#include <stdexcept>
#include "cubicspline.h"
#include <stdexcept>

#include "mydebug.h"

using namespace std;

cubicspline::cubicspline()
{
	mInitialized = false;
}

cubicspline::cubicspline(string inFilename)
{
        mIsNatural = true;
        attachFile(inFilename);
}

cubicspline::cubicspline(string inFilename, double inYp1, double inYpn)
{
        mIsNatural = false;
        mYp1 = inYp1;
        mYpn = inYpn;
        
	attachFile(inFilename);
}


cubicspline::cubicspline(istream& inStream)
{
	fromstream(inStream);
}

void
cubicspline::fromstream(istream& inStream)
{
	double x,y;
	
	try {
			while (inStream >> x) {
				inStream >> y;
				insertpoint(x,y);
			}
	} catch(...) {
		cerr << "error reading file";
	}

	initSplineTables();
}


cubicspline::~cubicspline()
{

}

void
cubicspline::attachFile(string inFilename)
{

	ifstream stream(inFilename.c_str());

	if (stream.fail() ) {
		throw runtime_error("unable to open file");
	}
	
	
	clearTables();
	
	fromstream(stream);

}

void
cubicspline::clearTables()
{
	mX.clear();
	mY.clear();
	mY2.clear();
}

void
cubicspline::insertpoint(double inX, double inY)
{

	// the complicated way for the STL to find something.
	vector<double>::iterator wherex = find_if(mX.begin(), mX.end(), bind1st(less<double>(), inX) );


	// We should enforce that no X value is used twice.  Algorithm would fail.
	// I don;'t know how to do this properly here, I'll do it in initSpline
/*	if (mX.at( distance(mX.begin(), wherex)) == inX) {
		cout << mX.at( wherex ) << " " << inX << endl;
		cerr << "Same X used twice.  Aborting.";
		throw;
	}
*/
	// Must insert vy first (par. 16.3.6) because wherex is invalid after insert()


	mY.insert(mY.begin() + distance(mX.begin(), wherex), inY);
	mX.insert(wherex, inX);

}

void
cubicspline::initSplineTables()
{
	cX = cY = NAN;
	mCacheEnabled = true;
        mCacheInitialized = false;

	// setting up the tables for cubic spline interpolation (second derivative zero at edges)
	// (Numerical recipes, section 3.3)
	size_t n = mY.size();
	mY2.resize(n);

	vector<double> u(n);
	
	// Check to see if same X used twice.  If so, abort.
	for (long k = 0; k < n - 1; ++k) {
		if (mX.at(k) == mX.at(k+1)) {
			cerr << "Same X used twice.  Check element [" << k << "]. Aborting.";
			throw logic_error("same X used twice in cubicspline.");
		}
	}

        if (mIsNatural) {
            mY2.at(0) = 0;
            u.at(0) = 0;
        } else {
            mY2.at(0) = -0.5;
            u.at(0) = (3./(mX.at(1)-mX.at(0))) * ((mY.at(1)-mY.at(0))/(mX.at(1)-mX.at(0))-mYp1);
        }

	for (size_t i = 1; i <= n - 2; i++) {

		double sig = (mX.at(i) - mX.at(i-1)) / (mX.at(i+1) - mX.at(i-1));
		double p = sig * mY2.at(i-1) + 2;
		mY2.at(i) = (sig - 1.)/p;
		u.at(i) =   (mY.at(i+1) - mY.at(i)) / (mX.at(i+1) - mX.at(i)) - (mY.at(i) - mY.at(i-1)) / (mX.at(i) - mX.at(i-1));
		u.at(i) = (6. * u.at(i) / (mX.at(i+1) - mX.at(i-1)) - sig * u.at(i-1)) / p;

	}

        double qn,un;
        
        if (mIsNatural) {
            un = 0.;
            qn = 0.;
        } else {
            qn = 0.5;
            un = (3./(mX.at(n-1)-mX.at(n-2))) * (mYpn - (mY.at(n-1)-mY.at(n-2))/(mX.at(n-1)-mX.at(n-2)));
        }

        mY2.at(n-1) = (un-qn*u.at(n - 2)) * (qn * mY2.at(n-2) + 1.);
	
	for (long k = n - 2; k >= 0; --k) {
		mY2.at(k) = mY2.at(k) * mY2.at(k+1) + u.at(k);
	}



	mInitialized = true;	
}

/*
double
cubicspline::y(double inX)
{
	// interpolation routine (Numerical recipes, section 3.3)
	if (!mInitialized) {
		cerr << "Spline table not initialized" << endl;
		throw;
	}
	if (mX.size() == 0) {
		cerr << "Spline table empty" << endl;
		throw;
	}
	
	if (inX == cX && mCacheEnabled == true) {
		return cY;
	}
	
	size_t lo = 0, hi = mX.size() - 1;	
	size_t k;
	
	while (hi - lo > 1) {
		k = (hi + lo) / 2;
		
		if (mX.at(k) > inX)
			hi = k;
		else
			lo = k;
				
	}
		
	double h;
	h = mX.at(hi) - mX.at(lo);

	
	if (h == 0.) {
		cerr << "Error in spline interpolation (bad table: duplicate x values)\n";
		throw;
	} else if (inX > mX.at(mX.size() - 1) || inX < mX.at(0) ) {
		cerr << "Error in cubicspline: out of range " << inX << 
				" not between " << mX.at(0) << " and " << mX.at(mX.size()-1) << endl;
		throw;
	}
	

	double a = (mX.at(hi) - inX) / h;
	double b = (inX - mX.at(lo)) / h;

	cout << "Debug ON\n";
		cout << mX.at(hi) << " " << inX << " " <<  h << endl;;
		cout << inX << " " << mX.at(lo) << " " << b << endl;
		cout << a << " " << b << endl;

		cout << "my2lo" << mY2.at(lo) << endl;
		cout << "my2hi" << mY2.at(hi) << endl;
	cout << "Debug OFF\n";

	cX = inX;
	cY = a * mY.at(lo) + b * mY.at(hi) + ( (a*a*a - a) * mY2.at(lo) + (b*b*b - b) * mY2.at(hi) ) * (h * h)/6.;	
	return cY;
}
*/
double
cubicspline::y(double inX)
{
	// interpolation routine (Numerical recipes, section 3.3)
	if (!mInitialized) {
		cerr << "Spline table not initialized" << endl;
		throw runtime_error("Spline table not initialized in cubicspline.");
	}
	if (mX.size() == 0) {
		cerr << "Spline table empty" << endl;
		throw runtime_error("Spline table empty in cubicspline.");
	}
	
	if (inX == cX && mCacheEnabled && mCacheInitialized) {
		return cY;
	}
	
	size_t lo = 0, hi = mX.size() - 1;	
	size_t k;
	
	while (hi - lo > 1) {
		k = (hi + lo) / 2;
		
		if (mX[k] > inX)
			hi = k;
		else
			lo = k;
				
	}
 
        double h;
	
	double xhi,xlo;
	xhi = mX[hi];
	xlo = mX[lo];
	
	h = xhi-xlo;

	
	if (h == 0.) {
		cerr << "Error in spline interpolation (bad table: duplicate x values)\n";
		throw runtime_error("duplicate x values in cubicspline.");
	} else if (inX > mX[mX.size() - 1] || inX < mX[0] ) {
		cerr << "Error in cubicspline: out of range " << inX << 
				" not between " << mX[0] << " and " << mX[mX.size()-1] << endl;
		throw runtime_error("out of range value in cubicspline.");
	}
	

	double a = (xhi - inX) / h;
	double b = (inX - xlo) / h;

/*	cout << "Debug ON\n";
		cout << mX.at(hi) << " " << inX << " " <<  h << endl;;
		cout << inX << " " << mX.at(lo) << " " << b << endl;
		cout << a << " " << b << endl;

		cout << "my2lo" << mY2.at(lo) << endl;
		cout << "my2hi" << mY2.at(hi) << endl;
	cout << "Debug OFF\n";
*/
	cX = inX;
	cY = a * mY[lo] + b * mY[hi] + ( (a*a*a - a) * mY2[lo] + (b*b*b - b) * mY2[hi] ) * (h * h)/6.;	
        mCacheInitialized = true;

	return cY;
}

valarray<double>
cubicspline::y(valarray<double>& inX)
{
	valarray<double> inY(inX.size());
	
	for (long i = 0; i < inY.size(); i++) {
		inY[i] = y(inX[i]);
	}
	
	return inY;
}


double
cubicspline::dydx(double inX)
{
	// interpolation routine (Numerical recipes, section 3.3)
	if (!mInitialized) {
		cerr << "Spline table not initialized" << endl;
		throw logic_error("Spline table not initialized in cubicspline.");
	}

	size_t lo = 0, hi = mX.size() - 1;	
	size_t k;
	
	while (hi - lo > 1) {
		k = (hi + lo) / 2;
		
		if (mX.at(k) > inX)
			hi = k;
		else
			lo = k;
				
	}
	
	double h = mX.at(hi) - mX.at(lo);
	
	if (h == 0.) {
		cerr << "Error in spline interpolation (bad table: duplicate x values)\n";
		throw runtime_error("duplicate x values in cubicspline.");
	}
	
	double a = (mX.at(hi) - inX) / h;
	double b = (inX - mX.at(lo)) / h;

	return (mY.at(hi) - mY.at(lo)) / h - (3*a*a - 1)/6 * h * mY2.at(lo) + (3*b*b - 1) / 6 * h * mY2.at(hi);
}

double
cubicspline::d2ydx2(double inX)
{
	// interpolation routine (Numerical recipes, section 3.3)
	if (!mInitialized) {
		cerr << "Spline table not initialized" << endl;
		throw runtime_error("Spline table not initialized in cubicspline.");

	}

	size_t lo = 0, hi = mX.size() - 1;	
	size_t k;
	
	while (hi - lo > 1) {
		k = (hi + lo) / 2;
		
		if (mX.at(k) > inX)
			hi = k;
		else
			lo = k;
				
	}
	
	double h = mX.at(hi) - mX.at(lo);
	
	if (h == 0.) {
		cerr << "Error in spline interpolation (bad table: duplicate x values)\n";
		throw runtime_error("duplicate x values in cubicspline.");

	}
	
	double a = (mX.at(hi) - inX) / h;
	double b = (inX - mX.at(lo)) / h;

	return a * mY2.at(lo) + b*mY2.at(hi);

}

